from datetime import (
    datetime,
    time,
)
from threading import Condition
from time import sleep

import freezegun
from assertpy import assert_that
from maintenance_util import (
    MaintenanceWindow,
    MaintenanceWindowManager,
)
from testdoubleutil import ForeverSleeper
from timeutil import (
    AwakableSleeper,
)


def test_run_func_on_start_by_default():
    func_stub = FuncStub()

    manager = MaintenanceWindowManager(
        maintenance_window=MaintenanceWindow(
            start_time=time(hour=4, minute=0, second=0, microsecond=0),
            end_time=time(hour=6, minute=0, second=0, microsecond=0),
        ),
        func=func_stub.func,
    )
    now = datetime.now().replace(hour=7)
    with freezegun.freeze_time(now):
        manager.start()

    assert_that(func_stub.called).is_true()

def test_do_not_run_func_on_start_when_requested_to():
    func_stub = FuncStub()

    manager = MaintenanceWindowManager(
        maintenance_window=MaintenanceWindow(
            start_time=time(hour=4, minute=0, second=0, microsecond=0),
            end_time=time(hour=6, minute=0, second=0, microsecond=0),
        ),
        func=func_stub.func,
        run_on_start=False,
    )
    now = datetime.now().replace(hour=7)
    with freezegun.freeze_time(now):
        manager.start()

    assert_that(func_stub.called).is_false()


def test_sleeps_until_maintenance_window_starts():
    sleep_condition = Condition()
    sleeper = NotifiableSleeper(condition=sleep_condition)

    window = MaintenanceWindow(
        start_time=time(hour=4, minute=0, second=0, microsecond=0),
        end_time=time(hour=6, minute=0, second=0, microsecond=0),
    )
    manager = MaintenanceWindowManager(
        maintenance_window=window,
        func=FuncStub().func,
        sleeper=sleeper,
    )
    now = datetime.now().replace(hour=7)
    with freezegun.freeze_time(now):
        with sleep_condition:
            manager.start()
            assert_that(sleep_condition.wait(timeout=1)).is_true()
            assert_that(sleeper.slept_for).is_equal_to(window.seconds_to_next())

def test_run_func_on_maintenance_window():
    func_condition = Condition()
    func_stub = NotifiableFunc(condition=func_condition)

    sleep_condition = Condition()
    sleeper = NotifiableSleeper(condition=sleep_condition)

    window = MaintenanceWindow(
        start_time=time(hour=4, minute=0, second=0, microsecond=0),
        end_time=time(hour=6, minute=0, second=0, microsecond=0),
    )
    manager = MaintenanceWindowManager(
        maintenance_window=window,
        func=func_stub.func,
        sleeper=sleeper,
    )
    now = datetime.now().replace(hour=7)
    with freezegun.freeze_time(now) as frozen_time:
        with sleep_condition:
            with func_condition:
                manager.start()
                for i in range(3):
                    assert_that(sleep_condition.wait(timeout=1)).is_true()
                    frozen_time.tick(delta=window.seconds_to_next())
                    assert_that(func_condition.wait(timeout=1)).is_true()

    manager.stop()

def test_stops_even_if_manager_is_sleeping():
    func_condition = Condition()
    func_stub = NotifiableFunc(condition=func_condition)

    sleep_condition = Condition()
    sleeper = ForeverSleeper(when_sleep_condition=sleep_condition)

    window = MaintenanceWindow(
        start_time=time(hour=4, minute=0, second=0, microsecond=0),
        end_time=time(hour=6, minute=0, second=0, microsecond=0),
    )
    manager = MaintenanceWindowManager(
        maintenance_window=window,
        func=func_stub.func,
        sleeper=sleeper,
    )
    now = datetime.now().replace(hour=7)
    with freezegun.freeze_time(now):
        with sleep_condition:
            with func_condition:
                manager.start()
                assert_that(sleep_condition.wait(timeout=1)).is_true()
                manager.stop()


class FuncStub(object):
    def __init__(self):
        self.called = False

    def func(self):
        self.called = True

class NotifiableFunc(object):
    def __init__(self, condition: Condition):
        self.condition = condition

    def func(self):
        with self.condition:
            self.condition.notify_all()

class NotifiableSleeper(AwakableSleeper):
    def __init__(self, condition: Condition):
        self.condition = condition
        self.slept_for = 0

    def sleep(self, secs):
        with self.condition:
            self.condition.notify_all()

        self.slept_for = secs
        sleep(0)

    def hibernate(self, secs):
        pass

    def wake_up(self):
        pass



from datetime import (
    datetime,
    time,
    timedelta,
)


class MaintenanceWindow(object):
    def __init__(
        self,
        start_time: time,
        end_time: time,
    ):
        self.start_time = start_time
        self.end_time = end_time

    def is_inside(self) -> bool:
        now = datetime.now().time()
        if self.start_time > self.end_time:
            return self.start_time <= now or now <= self.end_time
        else:
            return self.start_time <= now <= self.end_time

    def seconds_to_next(self):
        now = datetime.now()

        today_start = now.replace(
            hour=self.start_time.hour,
            minute=self.start_time.minute,
            second=self.start_time.second,
            microsecond=0,
        )

        if self.start_time > self.end_time:
            if self.is_inside() and now.time() > self.start_time:
                start_to_use = today_start + timedelta(days=1)
            else:
                start_to_use = today_start
        else:
            if self.is_inside():
                start_to_use = today_start + timedelta(days=1)
            elif now <= today_start:
                start_to_use = today_start
            else:
                start_to_use = today_start + timedelta(days=1)

        return (start_to_use - now).total_seconds()



import logging
from datetime import (
    datetime,
    timedelta,
)
from threading import Thread
from typing import (
    Callable,
    Optional,
)

from maintenance_util import MaintenanceWindow
from timeutil import (
    AwakableSleeper,
    ConditionSleeper,
)


logger = logging.getLogger(__name__)


class MaintenanceWindowManager(object):
    def __init__(
        self,
        maintenance_window: MaintenanceWindow,
        func: Callable[[], None],
        run_on_start: bool = True,
        sleeper: Optional[AwakableSleeper] = None
    ):
        self.maintenance_window = maintenance_window
        self.func = func
        self.run_on_start = run_on_start
        self.sleeper = sleeper or ConditionSleeper()
        self.thread: Optional[Thread] = None
        self.running = False


    def start(self):
        if self.run_on_start:
            logger.debug("Initializing manager, running func...")
            self.func()

        self.running = True
        self.thread = Thread(target=self._run, daemon=True)
        self.thread.start()

    def stop(self):
        self.running = False
        self.sleeper.wake_up()
        if self.thread:
            self.thread.join()

    def _run(self):
        while self.running:
            if self.maintenance_window.is_inside():
                logger.debug("Inside maintenance window, running func...")
                self.func()

            amount_to_sleep = self.maintenance_window.seconds_to_next()
            until = datetime.now() + timedelta(seconds=amount_to_sleep)
            logger.debug("Sleeping for %d seconds until %s", amount_to_sleep, until)
            self.sleeper.sleep(amount_to_sleep)



from datetime import (
    datetime,
    time,
)

import pytest

import freezegun
from assertpy import assert_that
from maintenance_util import MaintenanceWindow


@pytest.mark.parametrize(
    argnames=("start_hour", "end_hour", "current_hour", "result"),
    argvalues=(
        (4, 6, 5, True),
        (4, 6, 3, False),
        (4, 6, 7, False),
        (22, 4, 23, True),
        (22, 4, 3, True),
        (22, 4, 21, False),
        (22, 4, 5, False),
    ),
    ids=(
        "same day inside",
        "same day before",
        "same day after",
        "different days after start",
        "different days before end",
        "different days before start",
        "different days after end",
    )
)
def test_is_inside(start_hour, end_hour, current_hour, result):
    window = MaintenanceWindow(
        start_time=time(hour=start_hour, minute=0, second=0, microsecond=0),
        end_time=time(hour=end_hour, minute=0, second=0, microsecond=0),
    )
    now = datetime.now()

    with freezegun.freeze_time(now.replace(hour=current_hour)):
        assert_that(window.is_inside()).is_equal_to(result)


@pytest.mark.parametrize(
    argnames=("start_hour", "end_hour", "current_hour", "expected_hours"),
    argvalues=(
        (4, 6, 5, 23),
        (4, 6, 3, 1),
        (4, 6, 7, 21),
        (22, 4, 23, 23),
        (22, 4, 3, 19),
        (22, 4, 21, 1),
        (22, 4, 5, 17),
    ),
    ids=(
        "same day inside",
        "same day before",
        "same day after",
        "different days after start",
        "different days before end",
        "different days before start",
        "different days after end",
    )
)
def test_seconds_until_next(start_hour, end_hour, current_hour, expected_hours):
    window = MaintenanceWindow(
        start_time=time(hour=start_hour, minute=0, second=0, microsecond=0),
        end_time=time(hour=end_hour, minute=0, second=0, microsecond=0),
    )
    now = datetime.now().replace(minute=0, second=0, microsecond=0)

    with freezegun.freeze_time(now.replace(hour=current_hour)):
        assert_that(window.seconds_to_next()).is_equal_to(expected_hours * 3600)


def test_boundaries_are_inclusive():
    start_time = time(hour=4, minute=0, second=0, microsecond=0)
    end_time = time(hour=6, minute=0, second=0, microsecond=0)
    window = MaintenanceWindow(
        start_time=start_time,
        end_time=end_time,
    )
    now = datetime.now().replace(
        hour=start_time.hour,
        minute=start_time.minute,
        second=start_time.second,
        microsecond=start_time.microsecond,
    )
    with freezegun.freeze_time(now):
        assert_that(window.is_inside()).is_true()

    now = datetime.now().replace(
        hour=end_time.hour,
        minute=end_time.minute,
        second=end_time.second,
        microsecond=end_time.microsecond,
    )
    with freezegun.freeze_time(now):
        assert_that(window.is_inside()).is_true()
